%h1 Using an Intermediate Database for Queries

%p
  I have been using PostgreSQL for more than 15 years, and over that span of
  time it has always delivered a set of features that provide what I need to
  solve operational challenges.

%p
  Sometimes the challenge is that a production server is not authorized to use
  some features.  Here I will explore how to spin up a temporary database for
  running a query that depends on functionality the server warehousing the data
  does not have.
  %a{:href=>"https://www.postgresql.org/docs/current/plpython.html"} PL/Python
  is a example of a very useful extension that may not be installed or
  authorized in a given environment.  (It's an
  %q untrusted
  language, which means only superusers can define functions.)

%h2 Foreign Servers

%p
  %a{:href=>"https://www.postgresql.org/docs/current/postgres-fdw.html"}
  postgres_fdw
  provides a way to create a foreign table. If you are given a connection URL
  we need to split it into components

:codeblock
  :::ruby
  #!/usr/local/bin/ruby

  require 'uri'
  uri = URI("postgresql://query:XXX@db1/localharvest")
  components = [ :scheme, :user, :password, :host, :port, :path, :query, :fragment ]
  components.each do |part|
      puts "\#{part}=\#{uri.send(part)}"
  end

%p
  This can be fed into the rest of the script which creates the foreign servers

:codeblock
  :::sh
  #!/bin/sh
  eval `uri_to_pars.rb $0`

  psql -q $tmp_url <<SQL
    CREATE SCHEMA localharvest_remote;

    CREATE EXTENSION postgres_fdw;
    CREATE SERVER localharvest_query FOREIGN DATA WRAPPER postgres_fdw
      OPTIONS (host '${host}', dbname '${path##/}', fetch_size '30000');
    CREATE USER MAPPING FOR $USER SERVER localharvest_query
      OPTIONS (user '${user}', password '${password}');

    CREATE EXTENSION hstore;
    IMPORT FOREIGN SCHEMA public FROM SERVER localharvest_query INTO localharvest_remote;
  SQL
/ >

%h2 Queries From the Foreign Schema

%p
  Now we have a personal database that we can install extensions on.  The
  following makes
  %a{:href=>"https://www.nsc.liu.se/~kent/python-hostlist/"} python-hostlist
  accessible from within the database

:codeblock
  :::sql
  CREATE EXTENSION plpythonu;

  CREATE FUNCTION hostlist_collapse(names text[]) RETURNS text AS $$
  import hostlist
  return hostlist.collect_hostlist(names)
  $$ LANGUAGE plpythonu;

  CREATE FUNCTION hostlist_expand(names text) RETURNS text[] AS $$
  import hostlist
  return hostlist.expand_hostlist(names)
  $$ LANGUAGE plpythonu;

%p
  Now we can run queries using these utility functions

:codeblock
  :::sql
  SET search_path=localharvest_remote,public;

  SELECT hostlist_collapse(hosts)
  FROM hosts
  FROM machines;

  -- Stub time dependent functions for unit tests
  --
  CREATE SCHEMA unit_tests;
  SET search_path TO unit_tests, "postgres", public, pg_catalog;

  CREATE OR REPLACE FUNCTION now() RETURNS timestamp AS $$
  BEGIN
      RETURN clock_timestamp();
  END;
  $$ LANGUAGE plpgsql;


%p
  We could use the default schema
  = succeed ',' do
    %code public
  but using an alternate name space allows us to see what we've created.  If we
  want more performance we can also create materialized views to store a
  snapshot of the data locally.
